/*
 * Copyright (c) 2000 Marc Alexander Lehmann <pcg@goof.com>
 * 
 * Redistribution and use in source and binary forms, with or without modifica-
 * tion, are permitted provided that the following conditions are met:
 * 
 *   1.  Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 * 
 *   2.  Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 *   3.  The name of the author may not be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MER-
 * CHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPE-
 * CIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTH-
 * ERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/io.h>
#include <signal.h>

/* daemon mode. */
#define PERFORMANCE 96	/* cpu usage percentage at which to go into maximum performance mode. */
#define BATTERY     75  /* cpu usage percentage at which to enter battery optimized mode. */
#define MINSECS      5  /* enter battery optimized mode only when no activity this number of seconds. */

static int eax, ebx, ecx, edx, esi, edi;

static int opt_quiet;

static void
magic (int function, int state)
{
  asm("movl $-1, %%edi\n"
      "out %%al, (%%dx)\n"
      : "=a" (eax), "=b" (ebx), "=c" (ecx), "=d" (edx), "=S" (esi), "=D" (edi)
      : "a" (0x47534982), "b" (function), "c" (state), "d" (0xb2)
  );
}

static void
bailout (char *msg)
{
  fprintf (stderr, "FATAL: %s\n", msg);
  fprintf (stderr,
           "eax %08x\nebx %08x\necx %08x\nedx %08x\nesi %08x\nedi %08x\n",
           eax, ebx, ecx, edx, esi, edi);

  exit(77);
}

static int
get_state (void)
{
  magic (1, 1);

  if (eax != 0)
    bailout ("get state probably failed,\nplease report to <speedstep@plan9.de>, together with /proc/cpuinfo");

  /* 02020005 == on_ac_power, 02020004 == battery power */
  if (edi != 0x02020005 && edi != 0x02020004 && !opt_quiet)
    fprintf (stderr, "EDI == %08x != 02020005|02020004,\n"
                     "please report this to <speedstep@plan9.de>, together with /proc/cpuinfo\n"
                     "(suppress this notice with -q)\n", edi);

  if (!opt_quiet)
    printf ("current state: %s\n", ebx ? "battery optimized" : "maximum performance");

  return !!ebx;
}

static void
set_state (int slowdown)
{
  if (!opt_quiet)
    if (slowdown)
      printf ("switching to battery optimized mode\n");
    else
      printf ("switching to maximum performance mode\n");

  magic (2, slowdown);

  if (edi != 0)
    bailout ("EDI != 0: setting failed,\nplease report to <speedstep@plan9.de>, together with /proc/cpuinfo\n");

  if (ebx != slowdown)
    bailout ("could not switch power mode,\nplease report to <speedstep@plan9.de>, together with /proc/cpuinfo\n");
}

static void
usage (int rc)
{
  fprintf (stderr, "\n"
      "intel type 2 speedstep mode switching utility version " VERSION "\n"
      "\n"
      "Usage: speedstep [-qhbmg] [-s0] [-s1]\n"
      "         -h        this help\n"
      "         -q        quiet operation\n"
      "         -b | -s1  switch to battery optimized mode\n"
      "         -m | -s0  switch to maximum performance mode\n"
      "         -g        print current mode (default)\n"
      "         -G        exit immediately with the current mode as exit status\n"
      "\n"
      "         -d        go into daemon mode (forks)\n"
      "         -D        same as -d, but don't fork and log mode changes\n"
      "\n"
      "Please report bugs to <speedstep@plan9.de>\n"
      "http://speedstep.plan9.de/\n"
      "\n");
  exit (rc);
}

static int stat_fd;
static unsigned long stat_cpu, stat_noop;

static void
get_stat (void)
{
  char line[80];
  static unsigned long stat_user, stat_nice, stat_system, stat_idle;
  unsigned long n_user, n_nice, n_system, n_idle;

  lseek (stat_fd, 0, SEEK_SET);
  read (stat_fd, line, 79);
  line[79] = 0;

  if (4 != sscanf (line, "%*s %lu %lu %lu %lu", &n_user, &n_nice, &n_system, &n_idle))
    {
      fprintf (stderr, "unable to parse /proc/stat format (%s)\n", line);
      exit (77);
    }

  stat_cpu  = (n_user - stat_user) + (n_system - stat_system);
  stat_noop = stat_cpu + (n_nice - stat_nice) + (n_idle - stat_idle);

  stat_user   = n_user;
  stat_nice   = n_nice;
  stat_system = n_system;
  stat_idle   = n_idle;
}

static enum { MODE_P_DYNAMIC, MODE_B_DYNAMIC,
              MODE_P_STATIC,  MODE_B_STATIC } daemon_mode = MODE_B_DYNAMIC;

static int current_state;

#define SET_STATE(state) if (state != current_state) set_state ((current_state = state))
#define GET_STATE()      get_state ()

static void
daemon_loop (void)
{
  int p_hyst = 0;

  current_state = GET_STATE ();

  for (;;)
    switch (daemon_mode)
      {
        case MODE_P_DYNAMIC:
          SET_STATE (0);

          sleep (2); get_stat ();
          if (stat_cpu * 100 < BATTERY * stat_noop)
            {
              p_hyst += 2;
              if (p_hyst < MINSECS)
                break;

              daemon_mode = MODE_B_DYNAMIC;
            }

          p_hyst = 0;

          break;

        case MODE_B_DYNAMIC:
          SET_STATE (1);

          sleep (1); get_stat ();

          if (stat_cpu * 100 >= PERFORMANCE * stat_noop)
            daemon_mode = MODE_P_DYNAMIC;

          break;

        case MODE_P_STATIC:
          SET_STATE (0);
          pause ();
          break;

        case MODE_B_STATIC:
          SET_STATE (1);
          pause ();
          break;
      }
}

static void handle_signal (int signum)
{
  switch (signum)
    {
      case SIGHUP: 
        current_state = GET_STATE ();
        daemon_mode = current_state ? MODE_B_DYNAMIC : MODE_P_DYNAMIC;
        break;

      case SIGUSR1: daemon_mode = MODE_P_STATIC;  break;
      case SIGUSR2: daemon_mode = MODE_B_STATIC;  break;
    }
}

static void
install_sighands (void)
{
  struct sigaction sa;

  sa.sa_handler = handle_signal;
  sa.sa_flags = SA_RESTART;
  sigemptyset (&sa.sa_mask);

  sigaction (SIGHUP , &sa, 0);
  sigaction (SIGUSR1, &sa, 0);
  sigaction (SIGUSR2, &sa, 0);
}

static void
daemonize (void)
{
  stat_fd = open ("/proc/stat", O_RDONLY);

  if (!stat_fd)
    {
      perror ("/proc/stat");
      exit (1);
    }

  get_stat ();

  install_sighands ();

  chdir ("/");

  if (opt_quiet)
    {
      int i;

      for (i = 0; i <1023; i++)
        if (i != stat_fd)
          close (i);

      if (fork () != 0)
        exit (0);

      setsid ();
    }

  daemon_loop ();
}

int main(int argc, char *argv[])
{
  char *arg;

  ioperm (0xb2, 1, 1); /* we need port b2 */

  if (argc <= 1)
    {
      get_state ();
      exit (0);
    }
  else
    while ((arg = *++argv))
      if (arg[0] != '-')
        usage (1);
      else
        switch (arg[1]) {
          case 'q': opt_quiet++; break;
          case 'g': get_state (); break;
          case 'G': exit (get_state ()); break;
          case 'm': set_state (0); break;
          case 'b': set_state (1); break;

          case 's':
                    if (arg[2] == '0')
                      set_state (0);
                    else if (arg[2] == '1')
                      set_state (1);
                    else
                      usage (1);
                    break;

          case 'd': opt_quiet = 1; daemonize (); break;
          case 'D': daemonize (); break;

          case 'h':
          case '?': usage (0); break;

          default: usage(1); break;
        }

  return 0;
}

